package org.bitbucket.bradleysmithllc.etlunit;

import org.bitbucket.bradleysmithllc.etlunit.context.VariableContext;
import org.bitbucket.bradleysmithllc.etlunit.context.VariableContextImpl;
import org.bitbucket.bradleysmithllc.etlunit.parser.*;
import org.bitbucket.bradleysmithllc.etlunit.util.HashMapArrayList;
import org.bitbucket.bradleysmithllc.etlunit.util.MapList;
import org.bitbucket.bradleysmithllc.etlunit.util.NeedsTest;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

@NeedsTest
public class ClassBroadcasterImpl implements ClassBroadcaster
{
	private final ClassLocator locator;
	private final ClassListener listener;
	private final ClassDirector director;
	private final VariableContext context;
	private final Log log;
	private final MapLocal mapLocal;

	public ClassBroadcasterImpl(ClassLocator plocator, ClassDirector pdirector, ClassListener plistener, MapLocal mapLocal, Log log)
	{
		this(plocator, pdirector, plistener, mapLocal, log, new VariableContextImpl());
	}

	public ClassBroadcasterImpl(ClassLocator plocator, ClassDirector pdirector, ClassListener plistener, MapLocal mapLocal, Log log, VariableContext variableContext)
	{
		locator = plocator;
		listener = plistener;
		director = pdirector;
		this.log = log;
		context = variableContext;
		this.mapLocal = mapLocal;
	}

	@Override
	public int broadcast(StatusReporter statusReporter)
	{
		return broadcast0(statusReporter, false);
	}

	@Override
	public int broadcast(boolean scan_only)
	{
		return broadcast0(null, scan_only);
	}

	private int broadcast0(StatusReporter statusReporter, boolean scan_only)
	{
		if (statusReporter != null)
		{
			statusReporter.scanStarted();
		}

		int testCount = 0;

		StatusReporter.CompletionStatus classStatuses = new StatusReporter.CompletionStatus();

		director.beginBroadcast();

		while (locator.hasNext())
		{
			ETLTestClass cl = locator.next();

			mapLocal.setCurrentlyProcessingTestClass(cl);

			// we don't care about rejected or deferred results
			if (director.accept(cl) == ClassResponder.response_code.accept)
			{
				// create a subcontext for the class
				VariableContext classScope = context.createNestedScope();

				if (statusReporter != null)
				{
					statusReporter.testClassAccepted(cl);
				}

				List<ETLTestMethod> methodsToExecute = new ArrayList<ETLTestMethod>();

				List<ETLTestMethod> methods = cl.getTestMethods();

				Iterator<ETLTestMethod> mit = methods.iterator();

				// get a list of all methods before doing anything else
				while (mit.hasNext())
				{
					ETLTestMethod method = mit.next();

					if (director.accept(method) == ClassResponder.response_code.accept)
					{
						if (statusReporter != null)
						{
							statusReporter.testMethodAccepted(method);
						}

						testCount++;

						methodsToExecute.add(method);
					}
					else
					{
						log.debug("Test method not accepted:  " + method.getQualifiedName());
					}
				}

				if (methodsToExecute.size() > 0)
				{
					if (!scan_only)
					{
						try
						{
							listener.begin(cl, classScope);
						}
						catch (TestExecutionError err)
						{
							classStatuses.setError(err);
						}
						catch (TestAssertionFailure fail)
						{
							classStatuses.addFailure(fail);
						}
						catch (TestWarning warn)
						{
							classStatuses.addWarning(warn);
						}

						if (!classStatuses.hasFailures())
						{
							// broadcast the variables
							List<ETLTestVariable> varlist = cl.getClassVariables();
							Iterator<ETLTestVariable> varit = varlist.iterator();

							while (varit.hasNext())
							{
								ETLTestVariable var = varit.next();
								listener.declare(var, classScope);
							}
						}
					}

					try
					{
						// now broadcast all the before class methods.  Any failures here will break
						// the entire chain.
						List<ETLTestMethod> beforeClassMethods = cl.getBeforeClassMethods();
						Iterator<ETLTestMethod> bcit = beforeClassMethods.iterator();

						while (bcit.hasNext() && !classStatuses.hasFailures())
						{
							ETLTestMethod bcmethod = bcit.next();

							broadcast(bcmethod, classStatuses, scan_only, classScope);
						}
					}
					catch (TestExecutionError err)
					{
						classStatuses.setError(err);
					}

					mit = methodsToExecute.iterator();
					while (mit.hasNext())
					{
						ETLTestMethod method = mit.next();

						mapLocal.setCurrentlyProcessingTestMethod(method);

						List<ETLTestAnnotation> testAnnotList = method.getAnnotations("@Test");

						// we only care about the first one - all others are bogus.
						// There must be at least one or this method would not be a test method
						ETLTestAnnotation testAnnot = testAnnotList.get(0);

						StatusReporter.CompletionStatus methodStatus = new StatusReporter.CompletionStatus();

						if (statusReporter != null)
						{
							statusReporter.testBeginning(method);
						}

						if (classStatuses.hasFailures())
						{
							if (statusReporter != null)
							{
								statusReporter.testCompleted(method, classStatuses);
							}
						}
						else
						{
							try
							{
								ETLTestValueObject value = testAnnot.getValue();

								if (value != null)
								{
									ETLTestValueObject expect = value.query("expected-error-id");

									if (expect != null)
									{
										if (expect.getValueType() != ETLTestValueObject.value_type.quoted_string)
										{
											throw new TestExecutionError("Invalid expected-error-id - must be a string type",
													TestConstants.ERR_INVALID_TEST_ANNOTATION);
										}

										methodStatus.setExpectedErrorId(expect.getValueAsString());
									}

									expect = value.query("expected-failure-ids");

									if (expect != null)
									{
										if (expect.getValueType() != ETLTestValueObject.value_type.list)
										{
											throw new TestExecutionError("Invalid expected-failure-ids - must be a list type",
													TestConstants.ERR_INVALID_TEST_ANNOTATION);
										}

										methodStatus.addExpectedFailureIds(expect.getValueAsListOfStrings());
									}

									expect = value.query("expected-failure-id");

									if (expect != null)
									{
										if (expect.getValueType() != ETLTestValueObject.value_type.quoted_string)
										{
											throw new TestExecutionError("Invalid expected-failure-id - must be a string type",
													TestConstants.ERR_INVALID_TEST_ANNOTATION);
										}

										methodStatus.addExpectedFailureId(expect.getValueAsString());
									}

									expect = value.query("expected-warning-ids");

									if (expect != null)
									{
										if (expect.getValueType() != ETLTestValueObject.value_type.list)
										{
											throw new TestExecutionError("Invalid expected-warning-ids - must be a list type",
													TestConstants.ERR_INVALID_TEST_ANNOTATION);
										}

										methodStatus.addExpectedWarningIds(expect.getValueAsListOfStrings());
									}
								}

								runTest(method,
										cl.getBeforeTestMethods(),
										cl.getAfterTestMethods(),
										statusReporter,
										methodStatus,
										scan_only,
										classScope);
							}
							catch (TestExecutionError err)
							{
								methodStatus.setError(err);
							}
							catch (Throwable err)
							{
								log.severe("", err);
								methodStatus.setError(new TestExecutionError(err.toString(),
										TestConstants.ERR_UNCAUGHT_EXCEPTION,
										err));
							}

							// check to see if the any expected failures or errors occurred
							String expectedErrorId = methodStatus.getExpectedErrorId();

							if (expectedErrorId != null)
							{
								if (methodStatus.hasError())
								{
									if (!methodStatus.getError().getErrorId().equals(expectedErrorId))
									{
										methodStatus.addFailure(new TestAssertionFailure("Error '"
												+ expectedErrorId
												+ "' expected but '"
												+ methodStatus.getError().getErrorId()
												+ "' caught", TestConstants.FAIL_EXPECTED_EXCEPTION));
									}
									else
									{
										// this test passes
										methodStatus.setExpectedErrorOverride(true);
									}
								}
								else
								{
									methodStatus.addFailure(new TestAssertionFailure("Error '" + expectedErrorId + "' expected",
											TestConstants.FAIL_EXPECTED_EXCEPTION));
								}
							}

							List<String> failureIds = methodStatus.getExpectedFailureIds();

							if (failureIds.size() != 0)
							{
								List<String> assertionFailureIds = methodStatus.getAssertionFailureIds();
								if (failureIds.containsAll(assertionFailureIds) && assertionFailureIds.containsAll(failureIds))
								{
									// this test passes
									methodStatus.setExpectedFailureOverride(true);
								}
								else
								{
									methodStatus.addFailure(new TestAssertionFailure("Failure(s) "
											+ failureIds
											+ " expected but "
											+ assertionFailureIds
											+ " caught", TestConstants.FAIL_EXPECTED_ASSERTION));
								}
							}

							List<String> warningIds = methodStatus.getExpectedWarningIds();

							if (warningIds.size() != 0)
							{
								List<String> testWarningIds = methodStatus.getTestWarningIds();
								if (warningIds.containsAll(testWarningIds) && testWarningIds.containsAll(warningIds))
								{
									// nothing to do here because warnings can't fail a method except for expected warnings not happening
								}
								else
								{
									methodStatus.addFailure(new TestAssertionFailure("Warning(s) "
											+ warningIds
											+ " expected but "
											+ testWarningIds
											+ " caught", TestConstants.FAIL_EXPECTED_WARNING));
								}
							}

							if (statusReporter != null)
							{
								statusReporter.testCompleted(method, methodStatus);
							}
						}
					}
				}

				try
				{
					// now broadcast all the after class methods.  Any failures here will break
					// the entire chain.
					List<ETLTestMethod> afterClassMethods = cl.getAfterClassMethods();
					Iterator<ETLTestMethod> acit = afterClassMethods.iterator();

					while (acit.hasNext() && !classStatuses.hasFailures())
					{
						ETLTestMethod bcmethod = acit.next();

						broadcast(bcmethod, classStatuses, scan_only, classScope);
					}
				}
				catch (TestExecutionError err)
				{
					classStatuses.setError(err);
				}

				if (!scan_only)
				{
					try
					{
						listener.end(cl, classScope);
					}
					catch (TestExecutionError err)
					{
						classStatuses.setError(err);
					}
					catch (TestAssertionFailure err)
					{
						classStatuses.addFailure(err);
					}
					catch (TestWarning err)
					{
						classStatuses.addWarning(err);
					}
				}
			}
			else
			{
				log.info("Test class not accepted: " + cl.getQualifiedName());
			}
		}

		director.endBroadcast();

		if (statusReporter != null)
		{
			statusReporter.scanCompleted();
		}

		return testCount;
	}

	private void runTest(ETLTestMethod method, List<ETLTestMethod> beforeTestMethods, List<ETLTestMethod> afterTestMethods, StatusReporter statusReporter, StatusReporter.CompletionStatus methodStatus, boolean scan_only, VariableContext classScope)
			throws TestExecutionError
	{
		// create a nested scope
		VariableContext methodScope = classScope.createNestedScope();

		if (!scan_only)
		{
			try
			{
				listener.begin(method, methodScope);
			}
			catch (TestExecutionError err)
			{
				methodStatus.setError(err);
			}
			catch (TestAssertionFailure err)
			{
				methodStatus.addFailure(err);
			}
			catch (TestWarning err)
			{
				methodStatus.addWarning(err);
			}
		}

		// hit the before test methods
		Iterator<ETLTestMethod> bit = beforeTestMethods.iterator();

		while (bit.hasNext())
		{
			ETLTestMethod bmethod = bit.next();

			mapLocal.setCurrentlyProcessingMethod(bmethod);

			broadcast(bmethod, methodStatus, scan_only, methodScope);
		}

		if (!methodStatus.hasFailures())
		{
			mapLocal.setCurrentlyProcessingMethod(method);

			broadcast(method, methodStatus, scan_only, methodScope);
		}

		// hit the after test methods
		Iterator<ETLTestMethod> ait = afterTestMethods.iterator();

		while (ait.hasNext())
		{
			ETLTestMethod amethod = ait.next();

			mapLocal.setCurrentlyProcessingMethod(amethod);

			broadcast(amethod, methodStatus, scan_only, methodScope);
		}

		if (!scan_only)
		{
			try
			{
				listener.end(method, methodScope);
			}
			catch (TestExecutionError err)
			{
				methodStatus.setError(err);
			}
			catch (TestAssertionFailure err)
			{
				methodStatus.addFailure(err);
			}
			catch (TestWarning err)
			{
				methodStatus.addWarning(err);
			}
		}
	}

	private void broadcast(ETLTestMethod next, StatusReporter.CompletionStatus status, boolean scan_only, final VariableContext methodContext)
			throws TestExecutionError
	{
		List<ETLTestAnnotation> methodDefaults = next.getAnnotations("@OperationDefault");
		List<ETLTestAnnotation> classDefaults = next.getTestClass().getAnnotations("@OperationDefault");

		final MapList<String, ETLTestAnnotation> defaultMap = new HashMapArrayList<String, ETLTestAnnotation>();

		// grab the method defaults.  Overlay if a class default is present
		for (ETLTestAnnotation mdefault : methodDefaults)
		{
			Map<String, ETLTestValueObject> value = mdefault.getValue().getValueAsMap();

			// get the operation name
			ETLTestValueObject opName = value.get("operation");

			String opNameValueAsString = opName.getValueAsString();

			defaultMap.getOrCreate(opNameValueAsString).add(mdefault);
		}

		// grab the class defaults first.  Just drop these in
		for (ETLTestAnnotation cdefault : classDefaults)
		{
			Map<String, ETLTestValueObject> value = cdefault.getValue().getValueAsMap();

			// get the operation name
			ETLTestValueObject opName = value.get("operation");

			defaultMap.getOrCreate(opName.getValueAsString()).add(cdefault);
		}

		List<ETLTestOperation> ops = next.getOperations();

		Iterator<ETLTestOperation> it = ops.iterator();

		while (it.hasNext())
		{
			ETLTestOperation op = it.next();

			mapLocal.setCurrentlyProcessingTestOperation(op);

			if (director.accept(op) == ClassResponder.response_code.accept)
			{
				if (!scan_only)
				{
					try
					{
						ETLTestValueObject operands = resolveMergedOperationValue(defaultMap, op);

						final ExecutionContext econtext = new ExecutionContext()
						{
							@Override
							public void process(ETLTestOperation op, VariableContext vcontext)
									throws TestAssertionFailure, TestExecutionError, TestWarning
							{
								ETLTestValueObject operands = resolveMergedOperationValue(defaultMap, op);

								// handled or reject are both acceptable
								if (listener.process(op, operands, methodContext, this) == ClassResponder.action_code.defer)
								{
									throw new TestExecutionError("Listener could not handle operation: "
											+ op.getOperationName()
											+ ": "
											+ (operands != null ? operands.getJsonNode() : "{}"), TestConstants.ERR_INVALID_OPERATION);
								}
							}
						};

						listener.begin(op, operands, methodContext, econtext);

						try
						{
							if (listener.process(op, operands, methodContext, econtext) == ClassResponder.action_code.defer)
							{
								throw new TestExecutionError("Listener could not handle operation: " + op.getOperationName() + ": " + (
										operands != null
												? operands.getJsonNode()
												: "{}"), TestConstants.ERR_INVALID_OPERATION);
							}
						}
						finally
						{
							listener.end(op, operands, methodContext, econtext);
						}
					}
					catch (TestAssertionFailure asrt)
					{
						status.addFailure(asrt);
					}
					catch (TestWarning warn)
					{
						status.addWarning(warn);
					}
					catch (TestExecutionError thr)
					{
						throw thr;
					}
					catch (Throwable thr)
					{
						throw new TestExecutionError("Untrapped error", TestConstants.ERR_UNCAUGHT_EXCEPTION, thr);
					}
				}
			}
			else
			{
				log.info("Test operation not accepted:  " + op.getQualifiedName());
			}
		}
	}

	private ETLTestValueObject resolveMergedOperationValue(MapList<String, ETLTestAnnotation> defaultMap, ETLTestOperation op)
	{
		// TODO - lookup this operation against the map and merge defaults if found
		List<ETLTestAnnotation> opDefaults = defaultMap.get(op.getOperationName());

		ETLTestValueObject toperands = op.getOperands();

		if (opDefaults != null)
		{
			for (ETLTestAnnotation opdef : opDefaults)
			{
				Map<String, ETLTestValueObject> annotValueAsMap = opdef.getValue().getValueAsMap();

				ETLTestValueObject defaultValue = annotValueAsMap.get("defaultValue");

				if (toperands != null)
				{
					Map<String, ETLTestValueObject> toperandsValueAsMap = toperands.getValueAsMap();

					// verify the matchWhen and doNotMatchWhen conditions

					boolean shouldMerge = true;

					// run through matchWhen first
					if (annotValueAsMap.containsKey("matchWhen"))
					{
						for (String matchProperty : annotValueAsMap.get("matchWhen").getValueAsListOfStrings())
						{
							if (!toperandsValueAsMap.containsKey(matchProperty))
							{
								shouldMerge = false;
								break;
							}
						}
					}

					if (annotValueAsMap.containsKey("doNotMatchWhen"))
					{
						for (String matchProperty : annotValueAsMap.get("doNotMatchWhen").getValueAsListOfStrings())
						{
							if (toperandsValueAsMap.containsKey(matchProperty))
							{
								shouldMerge = false;
								break;
							}
						}
					}

					if (shouldMerge)
					{
						// this is a left merge.  Defaults can't override actual arguments
						toperands = toperands.merge(defaultValue, ETLTestValueObject.merge_type.left_merge);
					}
				}
				else
				{
					// if there is a matchwhen clause, then this always fails.  doNotMatchWhen is irrelevant
					if (!annotValueAsMap.containsKey("matchWhen"))
					{
						toperands = defaultValue;
					}
				}
			}
		}

		return toperands;
	}

	@Override
	public void reset()
	{
		locator.reset();
	}

	public VariableContext getVariableContext()
	{
		return context;
	}
}