package org.bitbucket.bradleysmithllc.etlunit.feature;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.github.fge.jackson.JsonLoader;
import com.github.fge.jsonschema.exceptions.ProcessingException;
import com.github.fge.jsonschema.main.JsonSchema;
import com.github.fge.jsonschema.main.JsonSchemaFactory;
import org.bitbucket.bradleysmithllc.etlunit.*;
import org.bitbucket.bradleysmithllc.etlunit.context.VariableContext;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestOperation;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestValueObject;
import org.bitbucket.bradleysmithllc.etlunit.util.JsonProcessingUtil;
import org.bitbucket.bradleysmithllc.etlunit.util.JsonSchemaResolver;
import org.bitbucket.bradleysmithllc.etlunit.util.StringUtils;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Type;

public class ResourceFeatureOperationSignature implements FeatureOperationSignature
{
	private final String id;
	private final FeatureOperation parent;
	private final boolean sub;
	private final ResourceFeatureMetaInfo resourceFeatureMetaInfo;
	private final ClassLoader loader;

	private final JsonSchema validator;
	private final JsonNode validatorNode;

	public ResourceFeatureOperationSignature(ResourceFeatureOperation resourceFeatureOperation, ResourceFeatureMetaInfo resourceFeatureMetaInfo, String sig, ClassLoader loader)
	{
		this.loader = loader;
		id = sig;
		parent = resourceFeatureOperation;
		this.resourceFeatureMetaInfo = resourceFeatureMetaInfo;
		sub = !parent.getName().equals(id);

		validatorNode = loadValidatorNode(loader);
		validator = loadValidator(validatorNode);
	}

	@Override
	public String getId()
	{
		return id;
	}

	@Override
	public JsonSchema getValidator()
	{
		return validator;
	}

	@Override
	public JsonNode getValidatorNode()
	{
		return validatorNode;
	}

	@Override
	public boolean isSubOperation()
	{
		return sub;
	}

	private JsonNode loadValidatorNode(ClassLoader loader)
	{
		String schema = resourceFeatureMetaInfo.getResource((isSubOperation() ? (parent.getName() + "." + id) : id) + ".validator.jsonSchema");

		if (schema != null)
		{
			try
			{
				return JsonSchemaResolver.resolveSchemaReference(JsonLoader.fromString(schema), loader);
			}
			catch (Exception e)
			{
				throw new RuntimeException("Invalid schema for operation signature [" + resourceFeatureMetaInfo.getReportableFeatureName()
						+ "].[" + id + "]", e);
			}
		}

		return null;
	}

	private JsonSchema loadValidator(JsonNode validatorNode)
	{
		if (validatorNode != null)
		{
			try
			{
				JsonSchemaFactory factory = JsonSchemaFactory.byDefault();

				return factory.getJsonSchema(validatorNode);
			}
			catch (ProcessingException e)
			{
				throw new RuntimeException("Invalid schema for operation signature [" + resourceFeatureMetaInfo.getReportableFeatureName()
						+ "].[" + id + "]: " + JsonProcessingUtil.getProcessingReport(e));
			}
			catch (Exception e)
			{
				throw new RuntimeException("Invalid schema for operation signature [" + resourceFeatureMetaInfo.getReportableFeatureName()
						+ "].[" + id + "]", e);
			}
		}

		return null;
	}

	@Override
	public action_code process(ETLTestOperation op, ETLTestValueObject parameters, VariableContext vcontext, ExecutionContext econtext)
			throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		// TODO - get application log
		return processIdealMethod(resourceFeatureMetaInfo.getDescribing().getListener(), resourceFeatureMetaInfo.getDescribing(), op, parameters, vcontext, econtext, resourceFeatureMetaInfo.getApplicationLog());
	}

	private action_code processIdealMethod(OperationProcessor listener, Feature feature, ETLTestOperation op, ETLTestValueObject obj_new, VariableContext context, ExecutionContext econtext, Log applicationLog)
			throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		Package p = feature.getClass().getPackage();
		String name = p.getName() + ".json." + StringUtils.makeUnderscoreName(feature.getFeatureName());

		applicationLog.info("Searching for handler for operation: '" + op.getOperationName() + "' with signature '" + id + "': " + (obj_new == null
				? ""
				: obj_new.getJsonNode()));

		String typeBase = name + "." + StringUtils.preparePropertyForPackageName(parent.getName()) + ".";

		if (isSubOperation())
		{
			typeBase += StringUtils.preparePropertyForPackageName(id) + ".";
		}

		typeBase += StringUtils.makeProperPropertyName(id);

		String typeName = typeBase + "Request";
		String interfaceName = typeBase + "Handler";

		// check if such types exist
		try
		{
			applicationLog.info("Searching for operation type : " + typeName);

			Class typeCl = feature.getClass().getClassLoader().loadClass(typeName);

			applicationLog.info("Resolved operation type : " + typeCl);

			try
			{
				applicationLog.info("Searching for interface type : " + interfaceName);
				Class interfaceCl = feature.getClass().getClassLoader().loadClass(interfaceName);

				applicationLog.info("Resolved operation processor type : " + interfaceCl);

				if (interfaceCl.isAssignableFrom(listener.getClass()))
				{
					applicationLog.info("Using operation processor interface");

					// iterate through the methods, and call the one which matches this signature
					for (Method method : interfaceCl.getMethods())
					{
						Type [] types = method.getGenericParameterTypes();

						if (
								types != null
										&&
										types.length == 5
										&&
										types[0].equals(typeCl)
										&&
										types[1].equals(ETLTestOperation.class)
										&&
										types[2].equals(ETLTestValueObject.class)
										&&
										types[3].equals(VariableContext.class)
										&&
										types[4].equals(ExecutionContext.class)
								)
						{
							// load the gson object
							// give gson a shot at filling it out

							ObjectMapper om = new ObjectMapper();

							String json = obj_new == null ? "{}" : obj_new.getJsonNode().toString();

							try {
								Object confObj = om.readValue(json, typeCl);

								Object o = method.invoke(listener, confObj, op, obj_new, context, econtext);

								return (action_code) o;
							} catch (IOException e) {
								throw new TestExecutionError(TestConstants.ERR_BAD_OPERANDS, "", e);
							}
						}
					}

					applicationLog.info("Feature exposes an interface type, but does not have the required process method");
				}
				else
				{
					applicationLog.info("Feature exposes an interface type, but does not implement it");
					return listener.process(op, obj_new, context, econtext);
				}
			}
			catch (ClassNotFoundException e)
			{
				applicationLog.info("Feature exposes an operation type, but not an interface");
			}
			catch (InvocationTargetException e)
			{
				if (e.getCause() instanceof TestAssertionFailure)
				{
					throw (TestAssertionFailure) e.getCause();
				}
				else if (e.getCause() instanceof TestWarning)
				{
					throw (TestWarning) e.getCause();
				}
				else if (e.getCause() instanceof TestExecutionError)
				{
					throw (TestExecutionError) e.getCause();
				}

				throw new TestExecutionError("", e);
			}
			catch (IllegalAccessException e)
			{
				applicationLog.info("Feature exposes an interface type, but does not expose the required process method");
			}
		}
		catch (ClassNotFoundException e)
		{
		}

		return listener.process(op, obj_new, context, econtext);
	}
}
